"""
Backend for django cache
"""
import socket
from django.core.cache import InvalidCacheBackendError
from django.core.cache.backends.memcached import PyLibMCCache
from django.utils.functional import cached_property
from .cluster_utils import get_cluster_info


class ElastiCache(PyLibMCCache):
    """
    backend for Amazon ElastiCache (memcached) with auto discovery mode
    it used pylibmc in binary mode
    """
    def __init__(self, server, params):
        self.update_params(params)
        super(ElastiCache, self).__init__(server, params)
        if len(self._servers) > 1:
            raise InvalidCacheBackendError(
                'ElastiCache should be configured with only one server '
                '(Configuration Endpoint)')
        if len(self._servers[0].split(':')) != 2:
            raise InvalidCacheBackendError(
                'Server configuration should be in format IP:port')

    def update_params(self, params):
        """
        update connection params to maximize performance
        """
        if not params.get('BINARY', True):
            raise Warning('To increase performance please use ElastiCache'
                          ' in binary mode')
        else:
            params['BINARY'] = True  # patch params, set binary mode
        if not 'OPTIONS' in params:
            # set special 'behaviors' pylibmc attributes
            params['OPTIONS'] = {
                'tcp_nodelay': True,
                'ketama': True
            }

    @cached_property
    def get_cluster_nodes(self):
        """
        return list with all nodes in cluster
        """
        server, port = self._servers[0].split(':')
        try:
            return get_cluster_info(server, port)['nodes']
        except (socket.gaierror, socket.timeout) as err:
            raise Exception('Cannot connect to cluster {} ({})'.format(
                self._servers[0], err
            ))

    @property
    def _cache(self):
        # PylibMC uses cache options as the 'behaviors' attribute.
        # It also needs to use threadlocals, because some versions of
        # PylibMC don't play well with the GIL.
        client = getattr(self._local, 'client', None)
        if client:
            return client

        client = self._lib.Client(self.get_cluster_nodes)
        if self._options:
            client.behaviors = self._options

        self._local.client = client

        return client



from setuptools import setup

import django_elasticache


setup(
    name='django-elasticache',
    version=django_elasticache.__version__,
    description='Django cache backend for Amazon ElastiCache (memcached)',
    long_description=open('README.rst').read(),
    author='Danil Gusev',
    platforms='any',
    author_email='danil.gusev@gmail.com',
    url='http://github.com/gusdan/django-elasticache',
    license='MIT',
    keywords='elasticache amazon cache pylibmc memcached aws',
    packages=['django_elasticache'],
    install_requires=['pylibmc', 'Django>=1.3'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Environment :: Web Environment :: Mozilla',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



"""
utils for discovery cluster
"""
from distutils.version import StrictVersion
import re
from telnetlib import Telnet


class WrongProtocolData(ValueError):
    """
    Exception for raising when we get something unexpected
    in telnet protocol
    """
    def __init__(self, cmd, response):
        super(WrongProtocolData, self).__init__(
            'Unexpected response {} for command {}'.format(response, cmd))


def get_cluster_info(host, port):
    """
    return dict with info about nodes in cluster and current version
    {
        'nodes': [
            'IP:port',
            'IP:port',
        ],
        'version': '1.4.4'
    }
    """
    client = Telnet(host, int(port))
    client.write('version\n')
    res = client.read_until('\r\n').strip()
    version_list = res.split(' ')
    if len(version_list) != 2 or version_list[0] != 'VERSION':
        raise WrongProtocolData('version', res)
    version = version_list[1]
    if StrictVersion(version) >= StrictVersion('1.4.14'):
        cmd = 'config get cluster\n'
    else:
        cmd = 'get AmazonElastiCache:cluster\n'
    client.write(cmd)
    res = client.read_until('\n\r\nEND\r\n')
    client.close()
    ls = filter(None, re.compile(r'\r?\n').split(res))
    if len(ls) != 4:
        raise WrongProtocolData(cmd, res)

    try:
        version = int(ls[1])
    except ValueError:
        raise WrongProtocolData(cmd, res)
    nodes = []
    try:
        for node in ls[2].split(' '):
            host, ip, port = node.split('|')
            nodes.append('{}:{}'.format(ip or host, port))
    except ValueError:
        raise WrongProtocolData(cmd, res)
    return {
        'version': version,
        'nodes': nodes
    }
